/*
 * Copyright: Nicholas P Wiebe (2008-2009) and Irmtraud M Meyer (2008-2009)
 * License: licensed under the GNU General Public License version 3 (GPLv3)
 * AlignedHelix.cpp
 */

#include "AlignedHelix.h"
#include "Alignment.h"
#include "Utilities.h"

#include <algorithm>
#include <cassert>
#include <math.h>
#include <sstream>

AlignedHelix::AlignedHelix(Alignment* align, const Helix & helix, int seqIndex, vector<int> & seq2AlignmentMap) {
	this->alignment = align;
	int alignedPos5, alignedPos3;

	for(int i = 0; i < helix.length; i++){
		alignedPos5 = seq2AlignmentMap[helix.pos5+i];
		alignedPos3 = seq2AlignmentMap[helix.pos3-i];

		bps.push_back(pair<int,int>(alignedPos5, alignedPos3));
	}
}


AlignedHelix::AlignedHelix(Alignment* align, int outerBpPos5, int outerBpPos3, int length){
	this->alignment = align;
	assert(outerBpPos5 >= 0);

	int pos5, pos3;
	for(int i = 0; i < length; i++){
		pos5 = outerBpPos5 + i;
		pos3 = outerBpPos3 - i;
		assert(pos3 > pos5);
		bps.push_back(pair<int,int>(pos5, pos3));
	}
}

AlignedHelix::~AlignedHelix() {

}

bool AlignedHelix::operator< (const AlignedHelix & rhs) const {
	for(int i = 0; i < min(bps.size(), rhs.bps.size()); ++i) {
		if (bps[i].first < rhs.bps[i].first) {
			return true;
		} else if (bps[i].first > rhs.bps[i].first) {
			return false;
		}
		
		if (bps[i].second < rhs.bps[i].second) {
			return true;
		} else if (bps[i].second > rhs.bps[i].second) {
			return false;
		}
	}
	return bps.size() > rhs.bps.size() ? true : false;
}


string AlignedHelix::dotBracket(unsigned int alignmentLength){
	string out = "";
	for(unsigned int i = 0; i < alignmentLength; i++){
		out += ".";
	}

	for(unsigned int i = 0; i < bps.size(); i++){
		out[bps[i].first] = '(';
		out[bps[i].second] = ')';
	}

	return out;
}

//added by Casper
string AlignedHelix::dotBracket(unsigned int alignmentLength, unsigned int start){
	string out = "";
	for(unsigned int i = start; i < alignmentLength; i++){
		out += ".";
	}

	for(unsigned int i = 0; i < bps.size(); i++){
		out[(bps[i].first)-start] = '(';
		out[(bps[i].second)-start] = ')';
	}

	return out;
}

bool AlignedHelix::insert(const Helix & helix, int seqIndex, vector<int> & seq2AlignmentMap){
	int alignedPos5, alignedPos3;

	//has to match exactly, including size!
	if(helix.length != (int)bps.size()){
		return false;
	}

	for(int i = 0; i < helix.length; i++){
		alignedPos5 = seq2AlignmentMap[helix.pos5+i];
		alignedPos3 = seq2AlignmentMap[helix.pos3-i];

		if(alignedPos5 != bps[i].first || alignedPos3 != bps[i].second){
			return false;
		}
	}
	return true;
}

double AlignedHelix::logLikeRatio(Tree * tree){
	int pos3, pos5;
	double loglike = 0;
	for(vector<pair<int, int> >::iterator it = bps.begin(); it != bps.end(); it++){
		pos5 = it->first;
		pos3 = it->second;
		loglike += alignment->logLikelihood(pos5,pos3, tree);
	}
	return loglike / bps.size();
}

double AlignedHelix::pairedLikelihood(Tree * tree){
	int pos3, pos5;
	double paired = 0;
	Alignment& alignRef = *alignment;
	for(vector<pair<int, int> >::iterator it = bps.begin(); it != bps.end(); it++){
		pos5 = it->first;
		pos3 = it->second;
		paired += tree->calcLogFelsPaired(alignRef, pos5, pos3);
	}
	return paired / bps.size();
}

double AlignedHelix::unpairedLikelihood(Tree * tree){
	int pos3, pos5;
	double unpaired = 0;
	Alignment& alignRef = *alignment;
	for(vector<pair<int, int> >::iterator it = bps.begin(); it != bps.end(); it++){
		pos5 = it->first;
		pos3 = it->second;
		unpaired += tree->calcLogFelsUnpaired(alignRef, pos5);
		unpaired += tree->calcLogFelsUnpaired(alignRef, pos3);
	}
	return unpaired / bps.size();
}

double AlignedHelix::canonicalBP(){
	int pos3, pos5;
	double canonicalBP = 0;

	for(vector<pair<int, int> >::iterator it = bps.begin(); it != bps.end(); it++){
		pos5 = it->first;
		pos3 = it->second;

		canonicalBP += alignment->canonicalBPPercent(pos5, pos3);
	}
	return canonicalBP / bps.size();
}

double AlignedHelix::covariance(){
	int pos3, pos5;
	double cov = 0;

	for(vector<pair<int, int> >::iterator it = bps.begin(); it != bps.end(); it++){
		pos5 = it->first;
		pos3 = it->second;

		cov += alignment->getCovariance(pos5, pos3);
	}
	return cov / bps.size();
}

double AlignedHelix::conservation(){
	int pos3, pos5;
	double conserve = 0;

	for(vector<pair<int, int> >::iterator it = bps.begin(); it != bps.end(); it++){
		pos5 = it->first;
		pos3 = it->second;

		conserve += alignment->getSeqCons(pos5);
		conserve += alignment->getSeqCons(pos3);
	}
	return conserve / (2*bps.size());
}

int AlignedHelix::length(){
	return bps.size();
}

//Modified by Casper to also allow bps to touch the boundary
HelixPosition AlignedHelix::positionFromBoundary(int boundaryIndex) {

	pair<int, int> outer_bp = bps.front();
	int left_outer_position = outer_bp.first;
	int right_outer_position = outer_bp.second;
	
	pair<int, int> inner_bp = bps.back();
	int left_inner_position = inner_bp.first;
	int right_inner_position = inner_bp.second; 

	if(bps.size()<Alignment::minStemLength){
		return INVALID_HELIX;
	}

	if (boundaryIndex > right_outer_position) {
		// boundary is completely right of the helix
		return LEFT_OF_BOUNDARY;	
	}
	
	else if (boundaryIndex <= left_outer_position) { //changed from < to <=
		// boundary is completely left of the helix
		return RIGHT_OF_BOUNDARY;
	}

//	else if (boundaryIndex > left_inner_position && boundaryIndex < right_inner_position) {
	else if (boundaryIndex > left_inner_position && boundaryIndex <= right_inner_position) {
		// boundary is between the left and right strands of the helix
		return OVERLAPPING_BOUNDARY;
	}

	else if((boundaryIndex<=right_outer_position && boundaryIndex>right_inner_position) || (boundaryIndex>left_outer_position && boundaryIndex<=left_inner_position) ){
		for(int i=0; i<bps.size(); i++){
			bps.erase(bps.begin());
			if(bps.size()<Alignment::minStemLength){
				return INVALID_HELIX;
			}

			left_outer_position=bps.front().first;
			right_outer_position=bps.front().second;
			left_inner_position = bps.back().first;
			right_inner_position = bps.back().second;
			if(!(boundaryIndex<=right_outer_position && boundaryIndex>right_inner_position) && !(boundaryIndex>left_outer_position && boundaryIndex<=left_inner_position)){
				if (boundaryIndex > right_outer_position) {
					// boundary is completely right of the helix
					return LEFT_OF_BOUNDARY;
				}
				else if (boundaryIndex <= left_outer_position) { //changed from < to <=
					// boundary is completely left of the helix
					return RIGHT_OF_BOUNDARY;
				}
			}
		}
	}
	else {
		// the boundary lies within the base pairs of the helix
		return INVALID_HELIX;
	}
}

string AlignedHelix::bpsString(){
	stringstream out;
	unsigned int i;
	for(i = 0; i < bps.size() - 1; i++){
		out << bps[i].first << ":" << bps[i].second << ",";
	}
	out << bps[i].first << ":" << bps[i].second;
	return out.str();
}

//added by Casper
string AlignedHelix::bpsString(unsigned int start){
	stringstream out;
	unsigned int i;
	for(i = 0; i < bps.size() - 1; i++){
		out << bps[i].first-start << ":" << bps[i].second-start<< ",";
	}
	out << bps[i].first-start << ":" << bps[i].second-start;
	return out.str();
}
